import numpy as np


class MovingAverage:
    def __init__(self):
        pass

    def filter_samples(self, samples, fs, span, sentinels=1):
        interval = 1000 / fs
        if span / 2 <= interval:
            return samples

        padding = 50
        padding = max(0, min(padding, len(samples) - (sentinels * 4)))
        padded = self.add_padding(samples, padding, sentinels)

        frames = int(span / (2 * interval))

        smoothed = np.convolve(padded, np.ones(frames) / frames, mode="same")
        return self.remove_padding(smoothed, padding)

    def add_padding(self, samples, padding, sentinels):
        pivot = sentinels
        left_padding = samples[pivot : pivot + padding]
        right_padding = samples[-sentinels - padding : -sentinels]

        reflected_left = samples[pivot] + (samples[pivot] - left_padding)
        reflected_right = samples[-sentinels - 1] + (
            samples[-sentinels - 1] - right_padding
        )

        padded = np.concatenate((reflected_left, samples, reflected_right))
        return padded

    def remove_padding(self, samples, padding):
        return samples[padding : len(samples) - padding]



import cv2
import numpy as np


class CalibratorPlane:
    def __init__(self):
        self.size = (0, 0)
        self.quad_image = [(0, 0), (0, 0), (0, 0), (0, 0)]
        self.origin = (0, 0)
        self.offset = (0, 0)

    def initialize(self, size, quad_image):
        self.size = size
        self.quad_image = quad_image
        self.origin = self.calculate_origin()

    def calculate_origin(self):
        # Assuming the origin is the bottom-left point of the quad
        return self.quad_image[3]

    def set_origin(self, p):
        # Set the origin of the calibration plane
        self.origin = p

    def update(self, quad_image):
        self.quad_image = quad_image

    def transform(self, point, origin):
        # Transform a point based on the calibration plane
        transformed_point = np.array(point) - np.array(origin) + np.array(self.offset)
        return tuple(transformed_point)

    def get_homography_matrix(self):
        # Calculate the homography matrix based on the user input quad
        src = np.array(self.quad_image, dtype=np.float32)
        dst = np.array([(0, 0), (self.size[0], 0), self.size, (0, self.size[1])], dtype=np.float32)

        return cv2.getPerspectiveTransform(src, dst)

    def transform_to_quad(self, point):
        # Transform a point to the user input quad
        homography_matrix = self.get_homography_matrix()
        transformed_point = cv2.perspectiveTransform(np.array([point], dtype=np.float32), np.linalg.inv(homography_matrix))
        return tuple(transformed_point[0][0])

class CalibratorLine:
    def __init__(self, quad):
        self.quad = quad  # Quadrilateral representing the calibration line
        self.origin = self.quad[0]  # Initial origin is the first point of the quad

    def set_origin(self, p):
        # Set the origin of the calibration line
        self.origin = p

    def transform(self, point):
        # Transform a point based on the calibration line
        # This is a simple example where the X-coordinate is transformed
        transformed_x = point[0] - self.origin[0]
        transformed_y = point[1]  # Y-coordinate remains unchanged
        return (transformed_x, transformed_y)
    


from datetime import datetime

import numpy as np


def get_time():
    # current date and time
    return datetime.now().strftime("%Y%m%d-%H%M%S")


def calculate_angle(a, o, b):
    # Reconstruct o,a,b from Point
    a = {
        "X": a.data["X"],
        "Y": a.data["Y"],
    }
    b = {
        "X": b.data["X"],
        "Y": b.data["Y"],
    }
    o = {
        "X": o.data["X"],
        "Y": o.data["Y"],
    }

    angle1 = np.arctan2(b["Y"] - o["Y"], b["X"] - o["X"])
    angle2 = np.arctan2(a["Y"] - o["Y"], a["X"] - o["X"])

    # Compute the angular difference in degrees
    ang = np.degrees(angle1 - angle2)

    # Adjust the result to be within [0, 360)
    if (ang < 0).any():
        ang[ang < 0] += 360
    return ang


# def calculate_angle(point1, point2, point3):
#     if len(point1.x) != len(point2.x) or len(point2.x) != len(point3.x):
#         raise ValueError("Input points must have the same length")

#     angle = np.arctan2(point3.y - point1.y, point3.x - point1.x) - np.arctan2(point2.y - point1.y, point2.x - point1.x)

#     return angle

# def calculate_angle(point1, point2, point3):
#     result = atan2(P3.y - P1.y, P3.x - P1.x) -
#                 atan2(P2.y - P1.y, P2.x - P1.x);
#     return result


# def buildComponent(num_frames, enum):
#     # return {each: np.full((num_frames), np.nan) for each in enum}
#     return {
#         "Raw X": np.full(num_frames, np.nan),
#         "Raw Y": np.full(num_frames, np.nan),
#         "X": np.full(num_frames, np.nan),
#         "Y": np.full(num_frames, np.nan),
#         "X Displacement": np.full(num_frames, np.nan),
#         "Y Displacement": np.full(num_frames, np.nan),
#         "Displacement": np.full(num_frames, np.nan),
#         "X Speed": np.full(num_frames, np.nan),
#         "Y Speed": np.full(num_frames, np.nan),
#         "Speed": np.full(num_frames, np.nan),
#         "X Acceleration": np.full(num_frames, np.nan),
#         "Y Acceleration": np.full(num_frames, np.nan),
#         "Acceleration": np.full(num_frames, np.nan),
#     }



BODY_JOINTS_MAP = {
    0: "Nose",
    1: "Left Eye",
    2: "Right Eye",
    3: "Left Ear",
    4: "Right Ear",
    5: "Left Shoulder",
    6: "Right Shoulder",
    7: "Left Elbow",
    8: "Right Elbow",
    9: "Left Wrist",
    10: "Right Wrist",
    11: "Left Hip",
    12: "Right Hip",
    13: "Left Knee",
    14: "Right Knee",
    15: "Left Ankle",
    16: "Right Ankle",
}



import numpy as np


class CircleFitter:
    @staticmethod
    def fit(traj):
        if traj.length < 3:
            return Circle.empty()

        # Least-squares circle fitting.
        rows = traj.length
        m = np.zeros((rows, 3))
        v = np.zeros((rows, 1))

        for i in range(rows):
            point = traj.coordinates(i)
            m[i, 0] = point[0]
            m[i, 1] = point[1]
            m[i, 2] = 1.0
            v[i, 0] = point[0] ** 2 + point[1] ** 2

        try:
            mt = m.transpose()
            b = np.dot(mt, m)
            c = np.dot(mt, v)
            z = np.linalg.solve(b, c)

            center = (z[0, 0] * 0.5, z[1, 0] * 0.5)
            radius = np.sqrt(z[2, 0] + (center[0] ** 2) + (center[1] ** 2))

            return Circle(center, radius)
        except np.linalg.LinAlgError:
            return Circle.empty()

class Circle:
    def __init__(self, center, radius):
        self.center = center
        self.radius = radius

    @staticmethod
    def empty():
        return Circle((0, 0), 0)



import math

import numpy as np


class ButterworthFilter:
    def __init__(self):
        self.a0, self.a1, self.a2 = 0, 0, 0
        self.b1, self.b2 = 0, 0
        self.x0, self.x1, self.x2 = 0, 0, 0
        self.y0, self.y1, self.y2 = 0, 0, 0
        self.correction_factor = 0

    def filter_samples(self, samples, fs, fc_tests):
        self.update_correction_factor(2)
        padding = 10
        padded = self.add_padding(samples, padding)

        filtered_results = []
        best_cutoff_index = -1

        for fc in np.linspace(0.5, fs / 2, fc_tests):
            filtered_padded = self.filter_samples_internal(padded, fs, fc)
            filtered = self.remove_padding(filtered_padded, padding)
            residuals = np.array(samples) - filtered
            dw = self.durbin_watson(residuals)

            if dw is not None and not math.isnan(dw):
                dw_normalized = abs(2 - dw) / 2
                result = FilteringResult(fc, filtered, dw_normalized)
                filtered_results.append(result)

                if (
                    best_cutoff_index == -1
                    or dw_normalized < filtered_results[best_cutoff_index].dw_normalized
                ):
                    best_cutoff_index = len(filtered_results) - 1

        return filtered_results, best_cutoff_index

    def update_correction_factor(self, passes):
        self.correction_factor = math.pow((math.pow(2, 1.0 / passes) - 1), 0.25)

    def update_coefficients(self, fs, fc):
        o = math.tan(math.pi * fc / fs) / self.correction_factor
        k1 = math.sqrt(2) * o
        k2 = o * o

        self.a0 = k2 / (1 + k1 + k2)
        self.a1 = 2 * self.a0
        self.a2 = self.a0

        k3 = 2 * self.a0 / k2
        self.b1 = -2 * self.a0 + k3
        self.b2 = 1 - self.a0 - self.a1 - self.a2 - self.b1

    def reset_values(self):
        self.x0, self.x1, self.x2 = 0, 0, 0
        self.y0, self.y1, self.y2 = 0, 0, 0

    def forward_pass(self, raw):
        self.reset_values()
        self.y1, self.x1 = raw[0], raw[0]
        self.y0, self.x0 = raw[1], raw[1]

        filtered = [self.y1, self.y0]
        for i in range(2, len(raw)):
            f = self.filter_sample(raw[i])
            filtered.append(f)
        return filtered

    def backward_pass(self, forward):
        self.reset_values()
        self.y1, self.x1 = forward[-1], forward[-1]
        self.y0, self.x0 = forward[-2], forward[-2]

        filtered = [self.y1, self.y0]
        for i in range(len(forward) - 3, -1, -1):
            f = self.filter_sample(forward[i])
            filtered.append(f)
        return filtered[::-1]

    def filter_samples_internal(self, samples, fs, fc):
        self.update_coefficients(fs, fc)
        forward = self.forward_pass(samples)
        backward = self.backward_pass(forward)
        return backward

    def filter_sample(self, sample):
        self.x2, self.x1, self.x0 = self.x1, self.x0, sample
        self.y2, self.y1 = self.y1, self.y0
        self.y0 = (
            self.a0 * self.x0
            + self.a1 * self.x1
            + self.a2 * self.x2
            + self.b1 * self.y1
            + self.b2 * self.y2
        )
        return self.y0

    def add_padding(self, samples, padding):
        padded = np.zeros(len(samples) + 2 * padding)
        for i in range(padding):
            padded[i] = samples[0] + samples[0] - samples[padding - i]
        for i in range(len(samples)):
            padded[padding + i] = samples[i]
        for i in range(padding):
            padded[padding + len(samples) + i] = (
                samples[-1] + samples[-1] - samples[len(samples) - 2 - i]
            )
        return padded

    def remove_padding(self, samples, padding):
        return samples[padding:-padding]

    def durbin_watson(self, residuals):
        diff = np.diff(residuals)
        numerator = sum(diff**2)
        denominator = sum(residuals**2)
        dw = numerator / denominator
        return dw


class FilteringResult:
    def __init__(self, fc, data, dw_normalized):
        self.fc = fc
        self.data = data
        self.dw_normalized = dw_normalized


class FilteredTrajectory:
    def __init__(self):
        self.length = 0
        self.can_filter = False
        self.times = []
        self.raw_xs = []
        self.raw_ys = []
        self.filter_result_xs = []
        self.x_cutoff_index = -1
        self.filter_result_ys = []
        self.y_cutoff_index = -1
        self.best_fit_circle = None

    def initialize(self, x, y, calibration_helper=None):
        self.length = len(x)
        self.raw_xs = x
        self.raw_ys = y

        self.can_filter = len(x) > 10

        if self.can_filter:
            # Framerate is set to 30 for this example, replace with your actual framerate
            framerate = 30

            filter = ButterworthFilter()

            tests = 10
            best_cutoff_index_x = 0
            self.filter_result_xs, best_cutoff_index_x = filter.filter_samples(
                self.raw_xs, framerate, tests
            )
            self.x_cutoff_index = best_cutoff_index_x

            best_cutoff_index_y = 0
            self.filter_result_ys, best_cutoff_index_y = filter.filter_samples(
                self.raw_ys, framerate, tests
            )
            self.y_cutoff_index = best_cutoff_index_y

        # I ADDED THIS 3 LINES TO GET THE FILTERED DATA
        # Check if the best cutoff indices are within valid range before accessing data
        if 0 <= self.x_cutoff_index < len(self.filter_result_xs):
            x_data = self.filter_result_xs[self.x_cutoff_index].data
        else:
            x_data = self.raw_xs

        if 0 <= self.y_cutoff_index < len(self.filter_result_ys):
            y_data = self.filter_result_ys[self.y_cutoff_index].data
        else:
            y_data = self.raw_ys
        return x_data, y_data

        # TODO: Calculate the best fit circle here if needed


if __name__ == "__main__":
    # Create an instance of FilteredTrajectory
    trajectory = FilteredTrajectory()

    # Initialize the trajectory with your data, e.g., a list of x and y values
    x = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
    y = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
    trajectory.initialize(x, y)

    x_data = trajectory.filter_result_xs[trajectory.x_cutoff_index].data
    y_data = trajectory.filter_result_ys[trajectory.y_cutoff_index].data



import json
import uuid
from pathlib import Path
from typing import Literal

import numpy as np
from matplotlib import pyplot as plt

from .point import Point
from .smoothing import MovingAverage
from .utils import calculate_angle, get_time


class Angle:
    def __init__(
        self,
        angle_name=None,
        dependencies=None,
        num_frames=0,
        fps=30,
        x_meter_per_pixel=None,
        y_meter_per_pixel=None,
    ):
        self.calibrationHelper = {
            "unit": "pixel",  # can be "pixel" or "meter
            "x_meter_per_pixel": 1.0
            if x_meter_per_pixel is None
            else float(x_meter_per_pixel),
            "y_meter_per_pixel": 1.0
            if y_meter_per_pixel is None
            else float(y_meter_per_pixel),
            "fps": 30 if fps is None else float(fps),
            "num_frames": 0 if num_frames is None else num_frames,
        }

        if dependencies:
            self.init_dependencies(dependencies)

        self.angle_name = f"{angle_name}" if angle_name else f"angle_{uuid.uuid4().hex}"
        self.data = {}
        self.smoothing_helper = None

    def init_dependencies(self, dependencies):
        self.dependencies = dependencies
        self.calibrationHelper = self.dependencies[0].calibrationHelper

    def preprocess_dependent_points(
        self,
    ):
        if self.dependencies:
            assert len(self.dependencies) == 3  # make sure got three points
            if self.dependencies:
                # check if all dependencies have the same num_frames, and have x and y
                for each in self.dependencies:
                    if np.isnan(each.data["X"]).all() or np.isnan(each.data["Y"]).all():
                        each.preprocess()
                self.calibrationHelper = self.dependencies[0].calibrationHelper
                assert all(
                    [
                        each.calibrationHelper["num_frames"]
                        == self.calibrationHelper["num_frames"]
                        for each in self.dependencies
                    ]
                )

    def preprocess(
        self,
        preprocess_interpolate_on=True,
        preprocess_filter_on=False,
        preprocess_smoothing_on=False,
    ):
        pass
        self.preprocess_dependent_points()

    def compute(self, postcalculate_filter_on=False, postcalculate_smoothing_on=False):
        self.computeAngle(
            self.dependencies[0], self.dependencies[1], self.dependencies[2]
        )
        self.computeAngleVelocity(smoothing_on=postcalculate_smoothing_on)
        self.computeAngleAcceleration(smoothing_on=postcalculate_smoothing_on)

    def computeAngle(self, point1, point2, point3):
        # compute angle
        self.data["Angle"] = calculate_angle(
            self.dependencies[0], self.dependencies[1], self.dependencies[2]
        )

    def computeAngleVelocity(self, smoothing_on=False):
        # compute angle velocity
        time = 1 / self.dependencies[0].calibrationHelper["fps"]
        self.data["Angle Velocity"] = (
            np.diff(self.data["Angle"], prepend=self.data["Angle"][0]) / time
        )

        if smoothing_on:
            self.data["Angle Velocity Smoothing"] = self.data["Angle Velocity"].copy()
            smoothing = self.smoothing_helper or MovingAverage()
            self.data["Angle Velocity Smoothing"] = smoothing.smooth(
                self.data["Angle Velocity Smoothing"]
            )
            # Update Angle Velocity
            self.data["Angle Velocity"] = self.data["Angle Velocity Smoothing"]

    def computeAngleAcceleration(self, smoothing_on=False):
        # compute angle acceleration
        time = 1 / self.dependencies[0].calibrationHelper["fps"]
        self.data["Angle Acceleration"] = (
            np.diff(self.data["Angle Velocity"], prepend=self.data["Angle Velocity"][0])
            / time
        )

        if smoothing_on:
            self.data["Angle Acceleration Smoothing"] = self.data[
                "Angle Acceleration"
            ].copy()
            smoothing = self.smoothing_helper or MovingAverage()
            self.data["Angle Acceleration Smoothing"] = smoothing.smooth(
                self.data["Angle Acceleration Smoothing"]
            )
            # Update Angle Acceleration
            self.data["Angle Acceleration"] = self.data["Angle Acceleration Smoothing"]

    def get_metrics(self):
        data = {
            "metadata": {
                "x_meter_per_pixel": self.dependencies[0].calibrationHelper[
                    "x_meter_per_pixel"
                ],
                "y_meter_per_pixel": self.dependencies[0].calibrationHelper[
                    "y_meter_per_pixel"
                ],
                "fps": self.dependencies[0].calibrationHelper["fps"],
                "unit": "degrees",
                "name": self.angle_name,
            },
            "metrics": {
                "O Point X": self.dependencies[0].data["X"].tolist(),
                "O Point Y": self.dependencies[0].data["Y"].tolist(),
                "A Point X": self.dependencies[1].data["X"].tolist(),
                "A Point Y": self.dependencies[1].data["Y"].tolist(),
                "B Point X": self.dependencies[2].data["X"].tolist(),
                "B Point Y": self.dependencies[2].data["Y"].tolist(),
                "Angle": self.data["Angle"].tolist(),
                "Angle Velocity": self.data["Angle Velocity"].tolist(),
                "Angle Velocity Smoothing"
                if "Angle Velocity Smoothing" in self.data
                else None: self.data["Angle Velocity Smoothing"].tolist()
                if "Angle Velocity Smoothing" in self.data
                else None,
                "Angle Acceleration": self.data["Angle Acceleration"].tolist(),
                "Angle Acceleration Smoothing"
                if "Angle Acceleration Smoothing" in self.data
                else None: self.data["Angle Acceleration Smoothing"].tolist()
                if "Angle Acceleration Smoothing" in self.data
                else None,
            },
            "metrics_info": {
                "O Point X": {
                    "unit": "pixel",
                },
                "O Point Y": {
                    "unit": "pixel",
                },
                "A Point X": {
                    "unit": "pixel",
                },
                "A Point Y": {
                    "unit": "pixel",
                },
                "B Point X": {
                    "unit": "pixel",
                },
                "B Point Y": {
                    "unit": "pixel",
                },
                "Angle": {
                    "unit": "degrees",
                },
                "Angle Velocity": {
                    "unit": "degrees/s",
                },
                "Angle Velocity Smoothing"
                if "Angle Velocity Smoothing" in self.data
                else None: {
                    "unit": "degrees/s",
                }
                if "Angle Velocity Smoothing" in self.data
                else None,
                "Angle Acceleration": {
                    "unit": "degrees/s^2",
                },
                "Angle Acceleration Smoothing"
                if "Angle Acceleration Smoothing" in self.data
                else None: {
                    "unit": "degrees/s^2",
                }
                if "Angle Acceleration Smoothing" in self.data
                else None,
            },
        }
        if None in data["metrics"]:
            del data["metrics"][None]
        if None in data["metrics_info"]:
            del data["metrics_info"][None]
        return data

    def export_metrics(self, dir):
        name = f"{self.angle_name}"

        # export json
        json_output = self.get_metrics()
        json_file = Path(dir) / get_time()
        json_file = json_file / f"{name}.json"
        json_file.parent.mkdir(parents=True, exist_ok=True)
        with open(json_file, "w") as f:
            json.dump(json_output, f, indent=2)

    def export(
        self,
        human_dir,
        mode: Literal["compact", "complete"] = "compact",
        frequency=30,
        meter_per_pixel=None,
        metrics_to_export=None,
    ):
        if self.dependencies is not None:
            data = {
                "Angle": self.data["Angle"].tolist(),
                "Angle Velocity": self.data["Angle Velocity"].tolist(),
                "Angle Acceleration": self.self.data["Angle Acceleration"].tolist(),
            }
            if mode == "compact":
                num_rows = 3  # Number of rows for the 3x3 grid
                # Adjust the figsize as needed
                fig, axes = plt.subplots(num_rows, figsize=(8, 6))

                for i, metric in enumerate(data):
                    ax = axes[i]
                    ax.plot(range(self.num_frames), data[metric])
                    ax.set_xlabel("Frame") if frequency is not None else ax.set_xlabel(
                        "Time (s)"
                    )
                    # You can adjust the unit accordingly
                    ax.set_ylabel("Degrees")
                    ax.set_title(metric)
                chart_file = Path(human_dir) / f"{self.angle_name}_compact.png"
                plt.tight_layout()
                plt.savefig(chart_file)
                plt.close()

            elif mode == "complete":
                for metric in data:
                    plt.figure(figsize=(8, 6))
                    plt.plot(range(self.num_frames), data[metric])
                    plt.xlabel("Frame") if frequency is not None else plt.xlabel(
                        "Time (s)"
                    )
                    # You can adjust the unit accordingly
                    plt.ylabel("Degrees")
                    plt.title(metric)
                    chart_file = Path(human_dir) / f"{self.angle_name}"
                    chart_file.mkdir(parents=True, exist_ok=True)
                    chart_file = (
                        chart_file / f"{metric.lower().replace(' ', '_')}_chart.png"
                    )
                    plt.savefig(chart_file)
                    plt.close()

            # export json
            json_output = self.get_metrics()
            json_file = Path(human_dir) / f"{self.angle_name}.json"
            with open(json_file, "w") as f:
                json.dump(json_output, f, indent=4)

    def loadFromJson(
        self,
        data,
        calibrationHelper=None,
        x_meter_per_pixel=None,
        y_meter_per_pixel=None,
        fps=None,
    ):
        # assert data shape
        assert len(data) == 3
        assert len(data["o"][0]) == 2
        self.dependencies = [
            Point(data=data["a"], num_frames=len(data["a"])),
            Point(data=data["o"], num_frames=len(data["o"])),
            Point(data=data["b"], num_frames=len(data["b"])),
        ]
        self.num_frames = self.dependencies[0].get_frames



import json
import uuid
from pathlib import Path
from typing import Any, Literal

import numpy as np
from matplotlib import pyplot as plt
from scipy import interpolate

from .constants import BODY_JOINTS_MAP
from .filtering import FilteredTrajectory
from .smoothing import MovingAverage
from .utils import get_time

PointMetrics = [
    "X",
    "Y",
    "X Displacement",
    "Y Displacement",
    "Displacement",
    "X Speed",
    "Y Speed",
    "Speed",
    "X Acceleration",
    "Y Acceleration",
    "Acceleration",
]


# So at first I let people initEmptyFrames so can easily update by using index. This is to ease if we know the length of data.
# But then we may not know the length of data, so I considering remove this thing. Either you create empty data and append, or you initialize with data. You have to add the empty frames if you initialize with data.
# In the end, if you use initWithData, I assume we already have the length. If you use appendPoints, you must update num_frames.


class Point:
    @property
    def get_frames(self):
        return self.calibrationHelper["num_frames"]

    def __init__(
        self,
        data=None,
        point_idx=None,
        point_name=None,
        num_frames=0,
        fps=30,
        x_meter_per_pixel=None,
        y_meter_per_pixel=None,
    ) -> None:
        # raw_ori_x and raw_ori_y are the original data
        # raw_x and raw_y are the interpolated data
        # x and y are the filtered data
        self.data = {}
        self.calibrationHelper = {}

        # TODO: The point can be dependent on other points, user must provide the function to calculate "Raw Ori X" and "Raw Ori Y" from the dependent points
        self.dependencies = []

        self.calibrationHelper = {
            "unit": "pixel",  # can be "pixel" or "meter
            "x_meter_per_pixel": 1.0
            if x_meter_per_pixel is None
            else float(x_meter_per_pixel),
            "y_meter_per_pixel": 1.0
            if y_meter_per_pixel is None
            else float(y_meter_per_pixel),
            "fps": 30 if fps is None else float(fps),
        }

        if data is not None:
            self.initWithData(data=data, fps=fps)
        else:
            self.initWithEmptyFrames(num_frames=num_frames, fps=fps)

        self.point_idx = point_idx
        self.point_name = (
            point_name
            if point_name
            else BODY_JOINTS_MAP[point_idx]
            if point_idx is not None
            else "point_" + uuid.uuid4().hex
        )
        self.filter = FilteredTrajectory()

    # SETUP
    def initWithEmptyFrames(self, num_frames, fps):
        self.data["Raw Ori X"] = np.full(num_frames, np.nan)
        self.data["Raw Ori Y"] = np.full(num_frames, np.nan)
        self.calibrationHelper["num_frames"] = num_frames

    def initWithData(self, data, fps=None):
        assert (
            len(data[0]) == 2
        ), (
            f"Expected 2 coordinates, got {len(data[0])}"
        )  # Assert data has two coordinates
        self.data["Raw Ori X"] = np.array(data)[:, 0]
        self.data["Raw Ori Y"] = np.array(data)[:, 1]
        self.calibrationHelper["num_frames"] = len(data)

    def __getitem__(self, __name: str) -> Any:
        return self.__dict__[__name]

    def preprocess_dependent_points(self, dependent_points):
        # TODO: Plot Raw Ori X and Raw Ori Y from the dependent points
        pass

    def preprocess(
        self,
        filter_threshold=None,
        preprocess_interpolate_on=True,
        preprocess_filter_on=True,
        preprocess_smoothing_on=False,
    ):
        # Find the first non-null data points from the beginning and end
        start_index, end_index = self.find_non_null_indices(self.data["Raw Ori X"])

        self.start_index = start_index if start_index is not None else 0
        self.end_index = (
            end_index
            if end_index is not None
            else self.calibrationHelper["num_frames"] - 1
        )

        # update X and Y
        self.data["X"] = self.data["Raw Ori X"]
        self.data["Y"] = self.data["Raw Ori Y"]

        if preprocess_interpolate_on:
            # Interpolate the data within the range defined by the first and last non-null data points
            self.data["Raw Ori X Interpolate"] = self.data["X"].copy()
            self.data["Raw Ori Y Interpolate"] = self.data["Y"].copy()
            self.data["Raw Ori X Interpolate"][
                self.start_index : self.end_index + 1
            ] = self.interpolate_missing(
                self.data["Raw Ori X Interpolate"][
                    self.start_index : self.end_index + 1
                ]
            )
            self.data["Raw Ori Y Interpolate"][
                self.start_index : self.end_index + 1
            ] = self.interpolate_missing(
                self.data["Raw Ori Y Interpolate"][
                    self.start_index : self.end_index + 1
                ]
            )

            # Update X and Y
            self.data["X"] = self.data["Raw Ori X Interpolate"]
            self.data["Y"] = self.data["Raw Ori Y Interpolate"]

        # Proceed with filtering if the number of frames is greater than 10
        can_filter = preprocess_filter_on and self.calibrationHelper["num_frames"] > 10

        if can_filter:
            self.data["Raw Ori X Filter"] = self.data["X"].copy()
            self.data["Raw Ori Y Filter"] = self.data["Y"].copy()
            # Apply filtering
            (
                self.data["Raw Ori X Filter"][self.start_index : self.end_index + 1],
                self.data["Raw Ori Y Filter"][self.start_index : self.end_index + 1],
            ) = self.filter.initialize(
                self.data["Raw Ori X Filter"][self.start_index : self.end_index + 1],
                self.data["Raw Ori Y Filter"][self.start_index : self.end_index + 1],
            )

            # Update X and Y
            self.data["X"] = self.data["Raw Ori X Filter"]
            self.data["Y"] = self.data["Raw Ori Y Filter"]

        if preprocess_smoothing_on:
            self.data["Raw Ori X Smooth"] = self.data["X"].copy()
            self.data["Raw Ori Y Smooth"] = self.data["Y"].copy()
            smoothing = MovingAverage()
            self.data["Raw Ori X Smooth"][
                self.start_index : self.end_index + 1
            ] = smoothing.filter_samples(
                samples=self.data["Raw Ori X Smooth"][
                    self.start_index : self.end_index + 1
                ],
                fs=self.calibrationHelper["fps"],
                span=40,
                sentinels=1,
            )
            self.data["Raw Ori Y Smooth"][
                self.start_index : self.end_index + 1
            ] = smoothing.filter_samples(
                samples=self.data["Raw Ori Y Smooth"][
                    self.start_index : self.end_index + 1
                ],
                fs=self.calibrationHelper["fps"],
                span=40,
                sentinels=1,
            )
            # Update X and Y
            self.data["X"] = self.data["Raw Ori X Smooth"]
            self.data["Y"] = self.data["Raw Ori Y Smooth"]

    def find_non_null_indices(self, data):
        # Find the first non-null data point from the beginning
        start_index = np.where(~np.isnan(data))[0]
        start_index = start_index[0] if start_index.size > 0 else None

        # Find the first non-null data point from the end
        end_index = np.where(~np.isnan(data))[0]
        end_index = end_index[-1] if end_index.size > 0 else None
        return int(start_index), int(end_index)

    def interpolate_missing(self, data):
        # Create an array of indices where values are not missing
        valid_indices = np.where(~np.isnan(data))[0]

        # Perform linear interpolation
        f = interpolate.interp1d(
            valid_indices, data[valid_indices], kind="linear", fill_value="extrapolate"
        )
        interpolated_data = f(np.arange(len(data)))

        return interpolated_data

    def compute(self, postcalculate_filter_on=False, postcalculate_smoothing_on=False):
        self.computeDisplacements()
        self.computeSpeeds(
            filter_on=postcalculate_filter_on, smoothing_on=postcalculate_smoothing_on
        )
        self.computeAccelerations(
            filter_on=postcalculate_filter_on, smoothing_on=postcalculate_smoothing_on
        )

        # Remove keys with all NaN values and replace with empty list
        for key, value in self.data.items():
            if np.isnan(value).all():
                self.data[key] = np.array([])

    def computeDisplacements(self):
        # Calculate displacement relative to the first frame
        x_displacement = (
            self.data["X"][self.start_index : self.end_index + 1]
            - self.data["X"][self.start_index]
        ) * self.calibrationHelper["x_meter_per_pixel"]

        y_displacement = (
            self.data["Y"][self.start_index : self.end_index + 1]
            - self.data["Y"][self.start_index]
        ) * self.calibrationHelper["y_meter_per_pixel"]

        self.data["X Displacement"] = self.data["X"].copy()
        self.data["Y Displacement"] = self.data["Y"].copy()
        self.data["Displacement"] = self.data["X"].copy()

        self.data["X Displacement"][
            self.start_index : self.end_index + 1
        ] = x_displacement
        self.data["Y Displacement"][
            self.start_index : self.end_index + 1
        ] = y_displacement
        # Calculate Displacement relative to the first frame
        self.data["Displacement"][self.start_index : self.end_index + 1] = np.sqrt(
            x_displacement**2 + y_displacement**2
        )

    def computeSpeeds(self, filter_on=False, smoothing_on=False):
        # X Speed Raw -> X Speed Filter -> X Speed Smooth -> X Speed
        time_diff = 1 / self.calibrationHelper["fps"]
        x_speed = (
            np.diff(
                self.data["X Displacement"][self.start_index : self.end_index + 1],
                prepend=self.data["X Displacement"][self.start_index],
            )
            / time_diff
        )
        y_speed = (
            np.diff(
                self.data["Y Displacement"][self.start_index : self.end_index + 1],
                prepend=self.data["Y Displacement"][self.start_index],
            )
            / time_diff
        )
        self.data["X Speed"] = self.data["X Displacement"].copy()
        self.data["Y Speed"] = self.data["Y Displacement"].copy()
        self.data["Speed"] = self.data["Displacement"].copy()

        self.data["X Speed"][self.start_index : self.end_index + 1] = x_speed
        self.data["Y Speed"][self.start_index : self.end_index + 1] = y_speed
        self.data["Speed"][self.start_index : self.end_index + 1] = np.sqrt(
            x_speed**2 + y_speed**2
        )
        self.data["X Speed Raw"] = self.data["X Speed"]
        self.data["Y Speed Raw"] = self.data["Y Speed"]
        self.data["Speed Raw"] = self.data["Speed"]

        if filter_on:
            self.data["X Speed Filter"] = self.data["X Speed"]
            self.data["Y Speed Filter"] = self.data["Y Speed"]
            # Apply filtering
            (
                self.data["X Speed Filter"][self.start_index : self.end_index + 1],
                self.data["Y Speed Filter"][self.start_index : self.end_index + 1],
            ) = self.filter.initialize(
                self.data["X Speed Filter"][self.start_index : self.end_index + 1],
                self.data["Y Speed Filter"][self.start_index : self.end_index + 1],
            )
            self.data["Speed Filter"] = np.sqrt(
                self.data["X Speed Filter"] ** 2 + self.data["Y Speed Filter"] ** 2
            )
            # Replace the speed with the filtered speed
            self.data["X Speed"] = self.data["X Speed Filter"]
            self.data["Y Speed"] = self.data["Y Speed Filter"]
            self.data["Speed"] = self.data["Speed Filter"]

        if smoothing_on:
            self.data["X Speed Smooth"] = self.data["X Speed"]
            self.data["Y Speed Smooth"] = self.data["Y Speed"]
            # Apply smoothing
            self.data["X Speed Smooth"][
                self.start_index : self.end_index + 1
            ] = MovingAverage.filter_samples(
                self.data["X Speed Smooth"][self.start_index : self.end_index + 1],
                self.calibrationHelper["fps"],
                40,
                1,
            )
            self.data["Y Speed Smooth"][
                self.start_index : self.end_index + 1
            ] = MovingAverage.filter_samples(
                self.data["Y Speed Smooth"][self.start_index : self.end_index + 1],
                self.calibrationHelper["fps"],
                40,
                1,
            )
            self.data["Speed Smooth"] = np.sqrt(
                self.data["X Speed Smooth"] ** 2 + self.data["Y Speed Smooth"] ** 2
            )

            # Replace the speed with the filtered speed
            self.data["X Speed"] = self.data["X Speed Smooth"]
            self.data["Y Speed"] = self.data["Y Speed Smooth"]
            self.data["Speed"] = self.data["Speed Smooth"]

    def computeAccelerations(self, filter_on=False, smoothing_on=True):
        # X Acceleration Raw -> X Acceleration Filter -> X Acceleration Smooth -> X Acceleration
        time_diff = 1 / self.calibrationHelper["fps"]
        x_accel = (
            np.diff(
                self.data["X Speed"][self.start_index : self.end_index + 1],
                prepend=self.data["X Speed"][self.start_index],
            )
            / time_diff
        )
        y_accel = (
            np.diff(
                self.data["Y Speed"][self.start_index : self.end_index + 1],
                prepend=self.data["Y Speed"][self.start_index],
            )
            / time_diff
        )

        self.data["X Acceleration"] = self.data["X Speed"].copy()
        self.data["Y Acceleration"] = self.data["Y Speed"].copy()
        self.data["Acceleration"] = self.data["Speed"].copy()

        self.data["X Acceleration"][self.start_index : self.end_index + 1] = x_accel
        self.data["Y Acceleration"][self.start_index : self.end_index + 1] = y_accel
        self.data["Acceleration"][self.start_index : self.end_index + 1] = np.sqrt(
            x_accel**2 + y_accel**2
        )
        self.data["X Acceleration Raw"] = self.data["X Acceleration"]
        self.data["Y Acceleration Raw"] = self.data["Y Acceleration"]
        self.data["Acceleration Raw"] = self.data["Acceleration"]

        if filter_on:
            self.data["X Acceleration Filter"] = self.data["X Acceleration"]
            self.data["Y Acceleration Filter"] = self.data["Y Acceleration"]

            # Apply filtering
            (
                self.data["X Acceleration Filter"][
                    self.start_index : self.end_index + 1
                ],
                self.data["Y Acceleration Filter"][
                    self.start_index : self.end_index + 1
                ],
            ) = self.filter.initialize(
                self.data["X Acceleration Filter"][
                    self.start_index : self.end_index + 1
                ],
                self.data["Y Acceleration Filter"][
                    self.start_index : self.end_index + 1
                ],
            )
            self.data["Acceleration Filter"] = np.sqrt(
                self.data["X Acceleration Filter"] ** 2
                + self.data["Y Acceleration Filter"] ** 2
            )

            # Replace the acceleration with the filtered acceleration
            self.data["X Acceleration"] = self.data["X Acceleration Filter"]
            self.data["Y Acceleration"] = self.data["Y Acceleration Filter"]
            self.data["Acceleration"] = self.data["Acceleration Filter"]

        if smoothing_on:
            self.data["X Acceleration Smooth"] = self.data["X Acceleration"]
            self.data["Y Acceleration Smooth"] = self.data["Y Acceleration"]

            # Apply smoothing
            self.data["X Acceleration"][
                self.start_index : self.end_index + 1
            ] = MovingAverage.filter_samples(
                self.data["X Acceleration"][self.start_index : self.end_index + 1],
                self.calibrationHelper["fps"],
                50,
                1,
            )
            self.data["Y Acceleration"][
                self.start_index : self.end_index + 1
            ] = MovingAverage.filter_samples(
                self.data["Y Acceleration"][self.start_index : self.end_index + 1],
                self.calibrationHelper["fps"],
                50,
                1,
            )
            self.data["Acceleration Smooth"] = np.sqrt(
                self.data["X Acceleration Smooth"] ** 2
                + self.data["Y Acceleration Smooth"] ** 2
            )

            # Replace the acceleration with the filtered acceleration
            self.data["X Acceleration"] = self.data["X Acceleration Smooth"]
            self.data["Y Acceleration"] = self.data["Y Acceleration Smooth"]
            self.data["Acceleration"] = self.data["Acceleration Smooth"]

    # SETTERS
    def append_empty_frames(self, frames=1):
        self.data["Raw Ori X"] = np.append(
            self.data["Raw Ori X"], np.full(frames, np.nan)
        )
        self.data["Raw Ori Y"] = np.append(
            self.data["Raw Ori Y"], np.full(frames, np.nan)
        )
        # update num_frames
        self.calibrationHelper["num_frames"] = len(self.data["Raw Ori X"])

    def append_points(self, data):
        assert len(data) == 2, f"Expected 2 coordinates, got {len(data)}"
        data = np.array(data)
        self.data["Raw Ori X"] = np.append(self.data["Raw Ori X"], data[0])
        self.data["Raw Ori Y"] = np.append(self.data["Raw Ori Y"], data[1])
        # update num_frames
        self.calibrationHelper["num_frames"] = len(self.data["Raw Ori X"])

    def set_frame_point(self, index, x, y):
        self.data["Raw Ori X"][index] = x
        self.data["Raw Ori Y"][index] = y
        return

    # GETTERS
    def get_metrics(self):
        data = {
            "metadata": self.get_metadata(),
            "metrics": self.get_metrics_data(),
            # { key: value.tolist() for key, value in self.data.items()},
            "metrics_info": self.get_metrics_info(),
            "time": [
                i / self.calibrationHelper["fps"]
                for i in range(self.calibrationHelper["num_frames"])
            ],
        }
        if None in data["metrics"]:
            del data["metrics"][None]
        if None in data["metrics_info"]:
            del data["metrics_info"][None]
        return data

    def get_frame_point(self, index):
        return self.data["Raw Ori X"][index], self.data["Raw Ori Y"][index]

    def get_metadata(self):
        return {
            "x_meter_per_pixel": self.calibrationHelper["x_meter_per_pixel"],
            "y_meter_per_pixel": self.calibrationHelper["y_meter_per_pixel"],
            "fps": self.calibrationHelper["fps"],
            "point_idx": self.point_idx,
            "num_frames": self.calibrationHelper["num_frames"],
            "name": self.point_name
            if self.point_name
            else BODY_JOINTS_MAP[self.point_idx]
            if self.point_idx is not None
            else None,
            "start_index": self.start_index,
            "end_index": self.end_index,
        }

    def get_metrics_data(self):
        data = {
            # coordinate
            "Raw Ori X": self.data["Raw Ori X"].tolist(),
            "Raw Ori Y": self.data["Raw Ori Y"].tolist(),
            "Raw Ori X Interpolate"
            if "Raw Ori X Interpolate" in self.data
            else None: self.data["Raw Ori X Interpolate"].tolist()
            if "Raw Ori X Interpolate" in self.data
            else None,
            "Raw Ori Y Interpolate"
            if "Raw Ori Y Interpolate" in self.data
            else None: self.data["Raw Ori Y Interpolate"].tolist()
            if "Raw Ori Y Interpolate" in self.data
            else None,
            "Raw Ori X Filter" if "Raw Ori X Filter" in self.data else None: self.data[
                "Raw Ori X Filter"
            ].tolist()
            if "Raw Ori X Filter" in self.data
            else None,
            "Raw Ori Y Filter" if "Raw Ori Y Filter" in self.data else None: self.data[
                "Raw Ori Y Filter"
            ].tolist()
            if "Raw Ori Y Filter" in self.data
            else None,
            "Raw Ori X Smooth" if "Raw Ori X Smooth" in self.data else None: self.data[
                "Raw Ori X Smooth"
            ].tolist()
            if "Raw Ori X Smooth" in self.data
            else None,
            "Raw Ori Y Smooth" if "Raw Ori Y Smooth" in self.data else None: self.data[
                "Raw Ori Y Smooth"
            ].tolist()
            if "Raw Ori Y Smooth" in self.data
            else None,
            "X": self.data["X"].tolist(),
            "Y": self.data["Y"].tolist(),
            # displacement
            "X Displacement": self.data["X Displacement"].tolist(),
            "Y Displacement": self.data["Y Displacement"].tolist(),
            "Displacement": self.data["Displacement"].tolist(),
            # speed
            "X Speed Raw": self.data["X Speed Raw"].tolist(),
            "Y Speed Raw": self.data["Y Speed Raw"].tolist(),
            "Speed Raw": self.data["Speed Raw"].tolist(),
            "X Speed Filter" if "X Speed Filter" in self.data else None: self.data[
                "X Speed Filter"
            ].tolist()
            if "X Speed Filter" in self.data
            else None,
            "Y Speed Filter" if "Y Speed Filter" in self.data else None: self.data[
                "Y Speed Filter"
            ].tolist()
            if "Y Speed Filter" in self.data
            else None,
            "Speed Filter" if "Speed Filter" in self.data else None: self.data[
                "Speed Filter"
            ].tolist()
            if "Speed Filter" in self.data
            else None,
            "Y Speed Smooth" if "Y Speed Smooth" in self.data else None: self.data[
                "Y Speed Smooth"
            ].tolist()
            if "Y Speed Smooth" in self.data
            else None,
            "X Speed Smooth" if "X Speed Smooth" in self.data else None: self.data[
                "X Speed Smooth"
            ].tolist()
            if "X Speed Smooth" in self.data
            else None,
            "Speed Smooth" if "Speed Smooth" in self.data else None: self.data[
                "Speed Smooth"
            ].tolist()
            if "Speed Smooth" in self.data
            else None,
            "X Speed": self.data["X Speed"].tolist(),
            "Y Speed": self.data["Y Speed"].tolist(),
            "Speed": self.data["Speed"].tolist(),
            # acceleration
            "X Acceleration Raw": self.data["X Acceleration Raw"].tolist(),
            "Y Acceleration Raw": self.data["Y Acceleration Raw"].tolist(),
            "Acceleration Raw": self.data["Acceleration Raw"].tolist(),
            "X Acceleration Filter"
            if "X Acceleration Filter" in self.data
            else None: self.data["X Acceleration Filter"].tolist()
            if "X Acceleration Filter" in self.data
            else None,
            "Y Acceleration Filter"
            if "Y Acceleration Filter" in self.data
            else None: self.data["Y Acceleration Filter"].tolist()
            if "Y Acceleration Filter" in self.data
            else None,
            "Acceleration Filter"
            if "Acceleration Filter" in self.data
            else None: self.data["Acceleration Filter"].tolist()
            if "Acceleration Filter" in self.data
            else None,
            "X Acceleration Smooth"
            if "X Acceleration Smooth" in self.data
            else None: self.data["X Acceleration Smooth"].tolist()
            if "X Acceleration Smooth" in self.data
            else None,
            "Y Acceleration Smooth"
            if "Y Acceleration Smooth" in self.data
            else None: self.data["Y Acceleration Smooth"].tolist()
            if "Y Acceleration Smooth" in self.data
            else None,
            "Acceleration Smooth"
            if "Acceleration Smooth" in self.data
            else None: self.data["Acceleration Smooth"].tolist()
            if "Acceleration Smooth" in self.data
            else None,
            "X Acceleration": self.data["X Acceleration"].tolist(),
            "Y Acceleration": self.data["Y Acceleration"].tolist(),
            "Acceleration": self.data["Acceleration"].tolist(),
        }
        if None in data:
            del data[None]
        return data

    def get_metrics_info(self):
        data = {
            # coordinate
            "Raw Ori X": {"Unit": "Pixel"},
            "Raw Ori Y": {"Unit": "Pixel"},
            "Raw Ori X Interpolate" if "Raw Ori X Interpolate" in self.data else None: {
                "Unit": "Pixel"
            },
            "Raw Ori Y Interpolate" if "Raw Ori Y Interpolate" in self.data else None: {
                "Unit": "Pixel"
            },
            "Raw Ori X Filter" if "Raw Ori X Filter" in self.data else None: {
                "Unit": "Pixel"
            },
            "Raw Ori Y Filter" if "Raw Ori Y Filter" in self.data else None: {
                "Unit": "Pixel"
            },
            "Raw Ori X Smooth" if "Raw Ori X Smooth" in self.data else None: {
                "Unit": "Pixel"
            },
            "Raw Ori Y Smooth" if "Raw Ori Y Smooth" in self.data else None: {
                "Unit": "Pixel"
            },
            "X": {"Unit": "Pixel"},
            "Y": {"Unit": "Pixel"},
            # displacement
            "X Displacement": {
                "Unit": "Meter"
                if self.calibrationHelper["x_meter_per_pixel"] is not None
                else "Pixel"
            },
            "Y Displacement": {
                "Unit": "Meter"
                if self.calibrationHelper["y_meter_per_pixel"] is not None
                else "Pixel"
            },
            "Displacement": {"Unit": "Meter"}
            if self.calibrationHelper["x_meter_per_pixel"] is not None
            and self.calibrationHelper["y_meter_per_pixel"] is not None
            else {"Unit": "Pixel"},
            # speed
            "X Speed": {"Unit": "Pixel/s"}
            if self.calibrationHelper["x_meter_per_pixel"] is None
            else {"Unit": "Meter/s"},
            "Y Speed": {
                "Unit": "Pixel/s"
                if self.calibrationHelper["y_meter_per_pixel"] is None
                else "Meter/s"
            },
            "Speed": {
                "Unit": "Pixel/s"
                if self.calibrationHelper["x_meter_per_pixel"] is None
                and self.calibrationHelper["y_meter_per_pixel"] is not None
                else "Meter/s"
            },
            # add in speed filter and speed smooth
            "X Speed Filter" if "X Speed Filter" in self.data else None: {
                "Unit": "Pixel/s"
            }
            if self.calibrationHelper["x_meter_per_pixel"] is None
            else {"Unit": "Meter/s"},
            "Y Speed Filter" if "Y Speed Filter" in self.data else None: {
                "Unit": "Pixel/s"
                if self.calibrationHelper["y_meter_per_pixel"] is None
                else "Meter/s"
            },
            "Speed Filter" if "Speed Filter" in self.data else None: {
                "Unit": "Pixel/s"
                if self.calibrationHelper["x_meter_per_pixel"] is None
                and self.calibrationHelper["y_meter_per_pixel"] is None
                else "Meter/s"
            },
            "X Speed Smooth" if "X Speed Smooth" in self.data else None: {
                "Unit": "Pixel/s"
            }
            if self.calibrationHelper["x_meter_per_pixel"] is None
            else {"Unit": "Meter/s"},
            "Y Speed Smooth" if "Y Speed Smooth" in self.data else None: {
                "Unit": "Pixel/s"
                if self.calibrationHelper["y_meter_per_pixel"] is None
                else "Meter/s"
            },
            "Speed Smooth" if "Speed Smooth" in self.data else None: {
                "Unit": "Pixel/s"
                if self.calibrationHelper["x_meter_per_pixel"] is None
                and self.calibrationHelper["y_meter_per_pixel"] is None
                else "Meter/s"
            },
            # acceleration
            "X Acceleration": {
                "Unit": "Pixel/s^2"
                if self.calibrationHelper["x_meter_per_pixel"] is None
                else "Meter/s^2"
            },
            "Y Acceleration": {
                "Unit": "Pixel/s^2"
                if self.calibrationHelper["y_meter_per_pixel"] is None
                else "Meter/s^2"
            },
            "Acceleration No Smooth"
            if "Acceleration No Smooth" in self.data
            else None: {
                "Unit": "Pixel/s^2"
                if self.calibrationHelper["x_meter_per_pixel"] is None
                and self.calibrationHelper["y_meter_per_pixel"] is None
                else "Meter/s^2"
            },
            "Acceleration": {
                "Unit": "Pixel/s^2"
                if self.calibrationHelper["x_meter_per_pixel"] is None
                and self.calibrationHelper["y_meter_per_pixel"] is None
                else "Meter/s^2"
            },
            # add in acceleration filter and acceleration smooth
            "X Acceleration Filter" if "X Acceleration Filter" in self.data else None: {
                "Unit": "Pixel/s^2"
                if self.calibrationHelper["x_meter_per_pixel"] is None
                else "Meter/s^2"
            },
            "Y Acceleration Filter" if "Y Acceleration Filter" in self.data else None: {
                "Unit": "Pixel/s^2"
                if self.calibrationHelper["y_meter_per_pixel"] is None
                else "Meter/s^2"
            },
            "Acceleration Filter" if "Acceleration Filter" in self.data else None: {
                "Unit": "Pixel/s^2"
                if self.calibrationHelper["x_meter_per_pixel"] is None
                and self.calibrationHelper["y_meter_per_pixel"] is None
                else "Meter/s^2"
            },
        }
        if None in data:
            del data[None]

    # EXPORTS
    def export_metrics(self, dir):
        if self.point_idx is not None:
            name = f"{self.point_idx}_{BODY_JOINTS_MAP[self.point_idx]}"
        else:
            name = self.point_name
        # export json
        json_output = self.get_metrics()
        json_file = Path(dir) / get_time()
        json_file = json_file / f"{name}.json"
        json_file.parent.mkdir(parents=True, exist_ok=True)
        with open(json_file, "w") as f:
            json.dump(json_output, f, indent=2)

    def export_diagrams(
        self,
        human_dir,
        mode: Literal['compact", "complete'] = "compact",
        frequency=30,
        meter_per_pixel=None,
        metrics_to_export=PointMetrics,
    ):
        # THINGS TO CONSIDER TO EXPORT JSON DATA
        # 1. frequency

        # THINGS TO CONSIDER TO EXPORT CHARTS
        # 1. what is the name of the diagram
        # 2. what is the x-ais unit? time or frame
        # 3. what is the y-axis unit? pixel or meter
        # 4. what is the dir to export? what is the png name?

        data = {
            "X Displacement": self.data["X Displacement"].tolist()
            if meter_per_pixel is None
            else (self.data["X Displacement"] * meter_per_pixel).tolist(),
            "Y Displacement": self.data["Y Displacement"].tolist()
            if meter_per_pixel is None
            else (self.data["Y Displacement"] * meter_per_pixel).tolist(),
            "X Coordinate": self.data["X"].tolist()
            if meter_per_pixel is None
            else (self.data["X"] * meter_per_pixel).tolist(),
            "Y Coordinate": self.data["Y"].tolist()
            if meter_per_pixel is None
            else (self.data["Y"] * meter_per_pixel).tolist(),
            "Displacement": self.data["Displacement"].tolist()
            if meter_per_pixel is None
            else (self.data["Displacement"] * meter_per_pixel).tolist(),
            "X Speed": self.data["X Speed"].tolist()
            if meter_per_pixel is None
            else (self.data["X Speed"] * meter_per_pixel).tolist(),
            "Y Speed": self.data["Y Speed"].tolist()
            if meter_per_pixel is None
            else (self.data["Y Speed"] * meter_per_pixel).tolist(),
            "Speed": self.data["Speed"].tolist()
            if meter_per_pixel is None
            else (self.data["Speed"] * meter_per_pixel).tolist(),
            "X Acceleration": self.data["X Acceleration"].tolist()
            if meter_per_pixel is None
            else (self.data["X Acceleration"] * meter_per_pixel).tolist(),
            "Y Acceleration": self.data["Y Acceleration"].tolist()
            if meter_per_pixel is None
            else (self.data["Y Acceleration"] * meter_per_pixel).tolist(),
            "Acceleration": self.data["Acceleration"].tolist()
            if meter_per_pixel is None
            else (self.data["Acceleration"] * meter_per_pixel).tolist(),
        }

        if self.point_idx is not None:
            name = f"{self.point_idx}_{BODY_JOINTS_MAP[self.point_idx]}"
        else:
            name = self.point_name
        if mode == "compact":
            num_metrics = len(metrics_to_export)
            num_cols = 3
            # unlimited rows
            num_rows = (
                num_metrics // num_cols + 1
                if num_metrics % num_cols != 0
                else num_metrics // num_cols
            )

            # Adjust the figsize as needed
            fig, axes = plt.subplots(num_rows, num_cols, figsize=(8, 6))

            for i, metric in enumerate(metrics_to_export):
                row = i // num_cols
                col = i % num_cols
                ax = axes[row, col]
                ax.plot(range(self.calibrationHelper["num_frames"]), data[metric])
                ax.set_xlabel("Frame") if frequency is not None else ax.set_xlabel(
                    "Time (s)"
                )
                ax.set_ylabel(
                    "Meter"
                ) if meter_per_pixel is not None else ax.set_ylabel("Pixel")
                ax.set_title(f"{metric}")  # Add chart title

            chart_filename = Path(human_dir)
            chart_filename.mkdir(parents=True, exist_ok=True)

            chart_filename = chart_filename / f"{name}_compact.png"
            plt.tight_layout()
            plt.savefig(chart_filename)
            plt.close()

        elif mode == "complete":
            for each in metrics_to_export:
                plt.figure(figsize=(8, 6))
                plt.plot(range(self.calibrationHelper["num_frames"]), data[each])
                plt.xlabel("Frame") if frequency is not None else plt.xlabel("Time (s)")
                plt.ylabel("Meter") if meter_per_pixel is not None else plt.ylabel(
                    "Pixel"
                )
                plt.title(f"{each}")
                chart_filename = (
                    Path(human_dir)
                    / f"{self.point_idx}_{BODY_JOINTS_MAP[self.point_idx]}"
                )
                chart_filename.mkdir(parents=True, exist_ok=True)
                chart_filename = (
                    chart_filename / f"{each.lower().replace(' ', '_')}_chart.png"
                )
                plt.savefig(chart_filename)
                plt.close()

        # export json
        json_output = self.get_metrics()
        json_file = Path(human_dir) / f"{name}.json"
        with open(json_file, "w") as f:
            json.dump(json_output, f, indent=4)



import csv
import json
from io import StringIO
from pathlib import Path
from typing import Any, Literal

from .human_profile import HumanProfile
from .utils import get_time


class Kinematics:
    def __init__(
        self,
        num_frames=None,
        save_dirs=None,
        raw_data=None,
        run_path=None,
        run_path_v2=True,
        x_meter_per_pixel=None,
        y_meter_per_pixel=None,
        fps=None,
    ) -> None:
        super().__init__()
        self.frequency = None
        self.profileSet: dict[str, HumanProfile] = {}
        self.source_path = None
        self.save_dirs = None
        self.num_frames = None
        self.calibrationHelper = None

        if raw_data is not None:
            self.setup(raw_data)
        elif run_path is not None and run_path_v2:
            self.setup_from_csv_v2(run_path)
        elif run_path is not None and not run_path_v2:
            self.setup_from_csv(run_path)

    def setup_from_csv(self, run_csv) -> None:
        if isinstance(run_csv, str):
            self.source_path = Path(run_csv)
            with open(run_csv, "r") as csv_file:
                csv_reader = csv.reader(csv_file)

                csv_list = list(csv_reader)
                accumulated_data = []  # Concat data from all frames
                for row in csv_list:
                    _, frame_data = row
                    frame_data = json.loads(frame_data)
                    accumulated_data.append(frame_data)

            # Broadcast accumulated data to each human profile
            for frame_idx, frame_data in enumerate(accumulated_data):
                # loop through each person in the frame
                for human_idx in frame_data:
                    if human_idx == "":
                        continue

                    # if the person does not exists, create a new one, and append idx empty frames, then process the frame
                    if human_idx not in self.profileSet:
                        self.profileSet[human_idx] = HumanProfile(human_idx=human_idx)
                        self.profileSet[human_idx].append_empty_frames(frames=frame_idx)
                        self.profileSet[human_idx].process_frame_and_append(
                            frame_data, human_idx=human_idx
                        )
                    # else if the person exists, just pass the whole frame into it, using human.process_frame(frame_data, human_idx=idx). Every human needs this to append empty frames
                    else:
                        self.profileSet[human_idx].process_frame_and_append(
                            frame_data, human_idx=human_idx
                        )
        else:
            raise TypeError("run_csv should be a string")

    def setup_from_csv_v2(self, run_csv):
        if isinstance(run_csv, str):
            self.source_path = Path(run_csv)
            with open(run_csv, "r") as csv_file:
                csv_reader = csv.reader(csv_file)

                csv_list = list(csv_reader)
                accumulated_data = []  # Concat data from all frames
                for row in csv_list:
                    _, frame_data = row
                    frame_data = json.loads(frame_data)
                    accumulated_data.append(frame_data)

                    # Keep Track of the human_idx and the frame data
                    for human_idx in frame_data:
                        if human_idx not in self.profileSet and human_idx != "":
                            self.profileSet[human_idx] = HumanProfile(
                                human_idx=human_idx
                            )

                    # Loop through each person, and pass all accumulated data to each person
                for human_idx, human in self.profileSet.items():
                    human.setup_from_data(accumulated_data, human_idx=human_idx)

    def setup_from_data(self, data):
        # This one, we already have everything formatted, as {human_idx: [frame_data * 17 * 2]}
        # Init three human profiles, and pass their respective data. Should use init_with_data function
        return

    def setup_from_stringio(self, run_csv) -> None:
        if isinstance(run_csv, StringIO):
            self.source_path = Path(__file__).resolve() / "temp.csv"
            csv_reader = csv.reader(run_csv)
            csv_list = list(csv_reader)
            self.num_frames = len(csv_list)
            for row in csv_list:
                frame_number, keypoints_data = row
                frame_number = int(frame_number)
                # Parse the JSON string
                keypoints_data = json.loads(keypoints_data)
                self.process_frame(frame_number - 1, keypoints_data)
        else:
            raise TypeError("run_csv should be a StringIO")

    def associate_human_profile(self, associate_human=[[]]):
        # rmb delete profile
        if len(associate_human) == 0 and len(associate_human[0]) == 0:
            return
        for each in associate_human:
            toRetain = each[0]
            self.profileSet[toRetain].associate_human_profile(
                [self.profileSet[i] for i in each]
            )
            for toRemove in each[1:]:
                self.profileSet.pop(toRemove)

    def compute(
        self,
        filter_threshold=None,
        preprocess_interpolate_on=False,
        preprocess_filter_on=False,
        preprocess_smoothing_on=False,
        postcalculate_filter_on=False,
        postcalculate_smoothing_on=False,
    ):
        for each, profile in self.profileSet.items():
            profile.compute(
                filter_threshold=filter_threshold,
                preprocess_interpolate_on=preprocess_interpolate_on,
                preprocess_filter_on=preprocess_filter_on,
                preprocess_smoothing_on=preprocess_smoothing_on,
                postcalculate_filter_on=postcalculate_filter_on,
                postcalculate_smoothing_on=postcalculate_smoothing_on,
            )

    def get_metrics(self):
        return {
            "metadata": {
                "num_frames": self.num_frames,
                "calibrationHelper": self.calibrationHelper,
            },
            "human_profiles": [each.get_metrics() for each in self.profileSet.values()],
        }

    def export(
        self,
        mode: Literal["compact, complete"],
        save_dirs=None,
        overwrite=False,
        meter_per_pixel=None,
    ) -> None:
        save_dirs = Path(self.source_path).parent if save_dirs is None else save_dirs
        for each, profile in self.profileSet.items():
            profile.export(
                save_dirs=save_dirs, mode=mode, meter_per_pixel=meter_per_pixel
            )

    def export_metrics(self, save_dirs=None) -> None:
        data = self.get_metrics()
        # save the json
        save_dirs = Path(self.source_path).parent if save_dirs is None else save_dirs
        json_file = Path(save_dirs) / get_time()
        json_file = json_file / "metrics.json"
        json_file.parent.mkdir(parents=True, exist_ok=True)
        with open(json_file, "w") as f:
            json.dump(data, f, indent=2)
        return

    def __call__(
        self,
        save=True,
        save_dirs=None,
        overwrite=True,
        mode="compact",
        meter_per_pixel=None,
        associate_human_profile=[],
        onlyHumanProfile=[],
        preprocess_interpolate_on=False,
        preprocess_filter_on=False,
        preprocess_smoothing_on=False,
        postcalculate_filter_on=False,
        postcalculate_smoothing_on=False,
    ) -> Any:
        # associate human profile
        if len(associate_human_profile) > 0 and len(associate_human_profile[0]) > 0:
            self.associate_human_profile(associate_human_profile)

        if len(onlyHumanProfile) > 0:
            self.profileSet = {
                each: self.profileSet[str(each)] for each in onlyHumanProfile
            }

        # compute
        self.compute(
            preprocess_interpolate_on=preprocess_interpolate_on,
            preprocess_filter_on=preprocess_filter_on,
            preprocess_smoothing_on=preprocess_smoothing_on,
            postcalculate_filter_on=postcalculate_filter_on,
            postcalculate_smoothing_on=postcalculate_smoothing_on,
        )

        if save:
            # self.export(save_dirs=save_dirs, overwrite=overwrite,
            #             mode=mode, meter_per_pixel=meter_per_pixel)
            self.export_metrics(save_dirs=save_dirs)

    # def _setup_from_csv_deprecated(self, run_csv) -> None:
    #     # Get the file name without extension using pathlib library
    #     if isinstance(run_csv, str):
    #         self.source_path = Path(run_csv)
    #         with open(run_csv, "r") as csv_file:
    #             csv_reader = csv.reader(csv_file)

    #             csv_list = list(csv_reader)
    #             self.num_frames = len(csv_list)
    #             for row in csv_list:
    #                 frame_number, keypoints_data = row
    #                 frame_number = int(frame_number)
    #                 # Parse the JSON string
    #                 keypoints_data = json.loads(keypoints_data)
    #                 self.process_frame(frame_number - 1, keypoints_data)
    #     else:
    #         raise TypeError("run_csv should be a string")

    # def process_frame(self, frame_number, keypoints_data):
    #     # make sure correct dimension
    #     for profile_id, profile_points in keypoints_data.items():
    #         if profile_id == "":  # cuz there's empty datapoints
    #             continue
    #         profile_id = int(profile_id)  # Convert profile_id to float
    #         if profile_id not in self.profileSet:
    #             self.profileSet[profile_id] = HumanProfile(
    #                 num_frames=self.num_frames, human_idx=profile_id
    #             )
    #         self.profileSet[profile_id].appendFrame(
    #             profile_points, frame_no=frame_number
    #         )



from typing import List

from pydantic import BaseModel


class Point(BaseModel):
    joint: List[List[float]]

    model_config = {
        "json_schema_extra": {
            "examples": [{"joint": [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]}]
        }
    }



from .calibrationPlane import CalibratorLine, CalibratorPlane


class CalibrationHelper:
    def __init__(self, image_size, calibrator_type='Line'):
        self.image_size = image_size
        self.calibrator_type = calibrator_type
        self.calibrator = None  # Initialize calibrator based on the calibrator_type

    def initialize_calibrator(self, quad):
        if self.calibrator_type == 'Line':
            self.calibrator = CalibratorLine(quad)
        elif self.calibrator_type == 'Plane':
            self.calibrator = CalibratorPlane(quad)

    def set_calibration_origin(self, p):
        if self.calibrator:
            self.calibrator.set_origin(p)

    def transform_point(self, point):
        if self.calibrator:
            return self.calibrator.transform(point)
        else:
            return point

    def get_length_text(self, p1, p2, precise=False, abbreviation=False):
        length = self.get_distance(self.transform_point(p1), self.transform_point(p2))
        return self.format_value(length, precise) + self.length_abbreviation(abbreviation)

    def format_value(self, value, precise=False):
        if precise:
            return '{:.2f}'.format(value)
        else:
            return '{:.0f}'.format(value)

    def length_abbreviation(self, abbreviation=False):
        if abbreviation:
            return " " + self.length_unit_abbreviation()
        else:
            return ""

    def length_unit_abbreviation(self):
        return "meter"



import csv
import json
import uuid
from pathlib import Path
from typing import Any, List

import numpy as np
import pandas as pd
from .angle import Angle
from .point import Point, PointMetrics
from .utils import get_time

# Two methods to init Human
# 1. We already know the len, in Point we can init empty frames and replace index
# 2. We don't know the len, in Point we can use append function


class HumanProfile:
    def __init__(
        self,
        human_idx=None,
        clean_data=None,
        frame_data=None,
        run_path=None,
        x_meter_per_pixel=None,
        y_meter_per_pixel=None,
        fps=None,
        body_format="coco",
    ) -> None:
        self.human_idx = human_idx
        self.name = f"{human_idx}" if human_idx else f"human_{uuid.uuid4().hex}"
        self.calibrationHelper = {
            "x_meter_per_pixel": 1.0
            if x_meter_per_pixel is None
            else float(x_meter_per_pixel),
            "y_meter_per_pixel": 1.0
            if y_meter_per_pixel is None
            else float(y_meter_per_pixel),
            "fps": 30 if fps is None else float(fps),
        }
        if body_format is None or body_format == "coco":
            self.body_joints = {
                i: Point(
                    point_idx=i,
                    x_meter_per_pixel=self.calibrationHelper["x_meter_per_pixel"],
                    y_meter_per_pixel=self.calibrationHelper["y_meter_per_pixel"],
                    fps=self.calibrationHelper["fps"],
                    num_frames=0,
                )
                for i in range(17)
            }
        if clean_data is not None:
            self.init_with_data(clean_data)
        elif frame_data is not None:
            self.process_frame_and_append(frame_data)
        elif run_path is not None:
            self.setup_from_csv(run_path)

        self.customRecord = {}

    def compute(
        self,
        filter_threshold=None,
        preprocess_interpolate_on=False,
        preprocess_filter_on=False,
        preprocess_smoothing_on=False,
        postcalculate_filter_on=False,
        postcalculate_smoothing_on=False,
    ) -> Any:
        for each, body_joint in self.body_joints.items():
            body_joint.preprocess(
                filter_threshold=filter_threshold,
                preprocess_filter_on=preprocess_filter_on,
                preprocess_interpolate_on=preprocess_interpolate_on,
                preprocess_smoothing_on=preprocess_smoothing_on,
            )
            body_joint.compute(
                postcalculate_filter_on=postcalculate_filter_on,
                postcalculate_smoothing_on=postcalculate_smoothing_on,
            )

        self.customRecord["Left Knee"] = Angle(
            angle_name="Left Knee",
            dependencies=[
                self.body_joints[11],
                self.body_joints[13],
                self.body_joints[15],
            ],
        )

        self.customRecord["Right Knee"] = Angle(
            angle_name="Right Knee",
            dependencies=[
                self.body_joints[12],
                self.body_joints[14],
                self.body_joints[16],
            ],
        )

        self.customRecord["Left Hip"] = Angle(
            angle_name="Left Hip",
            dependencies=[
                self.body_joints[5],
                self.body_joints[11],
                self.body_joints[13],
            ],
        )

        self.customRecord["Right Hip"] = Angle(
            angle_name="Right Hip",
            dependencies=[
                self.body_joints[6],
                self.body_joints[12],
                self.body_joints[14],
            ],
        )

        self.customRecord["Left Shoulder"] = Angle(
            angle_name="Left Shoulder",
            dependencies=[
                self.body_joints[7],
                self.body_joints[5],
                self.body_joints[11],
            ],
        )

        self.customRecord["Right Shoulder"] = Angle(
            angle_name="Right Shoulder",
            dependencies=[
                self.body_joints[8],
                self.body_joints[6],
                self.body_joints[12],
            ],
        )

        self.customRecord["Left Elbow"] = Angle(
            angle_name="Left Elbow",
            dependencies=[
                self.body_joints[5],
                self.body_joints[7],
                self.body_joints[9],
            ],
        )

        self.customRecord["Right Elbow"] = Angle(
            angle_name="Right Elbow",
            dependencies=[
                self.body_joints[6],
                self.body_joints[8],
                self.body_joints[10],
            ],
        )

        # TODO
        # 'center_of_mass' : Point()

        for each, customRecord in self.customRecord.items():
            customRecord.preprocess()
            customRecord.compute()

    def associate_human_profile(self, associate_human: List["HumanProfile"]):
        assert self == associate_human[0]
        for frame in range(self.get_num_frames):
            x, y = self.body_joints[0].get_frame_point(frame)
            if x and y:
                pass
            else:
                for each in associate_human[1:]:
                    x, y = each.body_joints[0].get_frame_point(frame)
                    if x and y:
                        for joint_idx, each_joint in self.body_joints.items():
                            each_joint.set_frame_point(frame, x, y)
                        break

    # GETTERS
    @property
    def get_num_frames(self):
        n = len(self.body_joints[0].data["Raw Ori X"])
        assert n == len(
            self.body_joints[0].data["Raw Ori Y"]
        ), "Inconsistent data length"
        return n

    def get_joint_frame_point(self, frame):
        return self.body_joints[0].get_frame_point(frame)

    def get_body_joints_data(self):
        data = {}
        for each, body_joint in self.body_joints.items():
            data[each] = {key: item.tolist() for key, item in body_joint.data}

    def get_metadata(self):
        return (
            {
                "x_meter_per_pixel": self.calibrationHelper["x_meter_per_pixel"],
                "y_meter_per_pixel": self.calibrationHelper["y_meter_per_pixel"],
                "fps": self.calibrationHelper["fps"],
                "human_idx": self.human_idx,
                "name": self.name,
                "start_index": self.body_joints[0].get_metadata()["start_index"],
                "end_index": self.body_joints[0].get_metadata()["end_index"],
            },
        )

    def get_metrics(self):
        data = {
            "metadata": self.get_metadata(),
            "body_joints_metrics": {
                # key: {
                #     "metrics": joint.get_metrics()["metrics"],
                #     "metadata": {
                #         "name": joint.get_metrics()["metadata"]["name"],
                #         "start_index": joint.get_metrics()["metadata"]["start_index"],
                #         "end_index": joint.get_metrics()["metadata"]["start_index"],
                #     },
                # }
                key: joint.get_metrics_data()
                for key, joint in self.body_joints.items()
            },
            "body_joints_metrics_info": self.body_joints[0].get_metrics_info(),
            "custom_metrics": {
                key: metric.get_metrics()["metrics"]
                for key, metric in self.customRecord.items()
            },
            "custom_metrics_info": self.customRecord["Left Knee"].get_metrics()[
                "metrics_info"
            ],
            "time": [
                i / self.calibrationHelper["fps"] for i in range(self.get_num_frames)
            ],
        }

        return data

    # SETTERS AND IMPORT
    def setup_from_csv(self, run_csv, human_idx=1):
        if isinstance(run_csv, str):
            self.source_path = Path(run_csv)
            with open(run_csv, "r") as csv_file:
                csv_reader = csv.reader(csv_file)

                csv_list = list(csv_reader)

                accumulated_data = []  # Concat data of only one human from all frames
                for row in csv_list:
                    frame_number, frame_data = row
                    frame_number = int(frame_number)
                    # Parse the JSON string
                    frame_data = json.loads(frame_data)

                    processed_frame_data = self.process_frame(
                        frame_data, human_idx=self.human_idx or human_idx
                    )  # TODO: This is hardcoded to 1, need to change to self.human_idx or the parameter

                    accumulated_data.append(processed_frame_data)
                accumulated_data = np.array(accumulated_data, dtype=float)
                self.init_with_data(accumulated_data)

    def setup_from_data(self, data, human_idx=1):
        # Exact same as setup_from_csv, but we don't need to parse the JSON string, already formatted as [frames][dict:[17 joints][x, y]]
        # It's taking all frames and all humans, but it only keeps the data of one human
        accumulated_data = []
        for frame_data in data:
            processed_frame_data = self.process_frame(
                frame_data, human_idx=self.human_idx or human_idx
            )
            accumulated_data.append(processed_frame_data)
        accumulated_data = np.array(accumulated_data, dtype=float)
        self.init_with_data(accumulated_data)

    def process_frame(self, frame_data, human_idx=1):
        if isinstance(human_idx, int):
            human_idx = str(human_idx)
        if human_idx in frame_data:
            return frame_data[human_idx]
        else:
            return [[None, None] for _ in range(len(self.body_joints))]

    def process_frame_and_append(self, frame_data, human_idx=1):
        if isinstance(human_idx, int):
            human_idx = str(human_idx)

        if human_idx in frame_data:
            self.process_profile_and_append(frame_data[human_idx])
            return frame_data[human_idx]
        else:
            self.append_empty_frames(frames=1)
            return [[None, None] for _ in range(len(self.body_joints))]

    def process_profile_and_append(self, points: list[list[float]]) -> None:
        assert len(points) == len(
            self.body_joints
        ), f"Expected {len(self.body_joints)} points, got {len(points)}"

        for each, body_joint in self.body_joints.items():
            body_joint.append_points(points[each])

    def init_with_data(self, body_joints_data):
        # body_joints_data shape: [frames][17 joints][x, y]
        if isinstance(body_joints_data, list):
            body_joints_data = np.array(body_joints_data, dtype=float)
        self.body_joints = {
            i: Point(
                point_idx=i,
                fps=self.calibrationHelper["fps"],
                data=body_joints_data[:, i],
            )
            for i in range(17)
        }

    def append_empty_frames(self, frames=1):
        for each, body_joint in self.body_joints.items():
            body_joint.append_empty_frames(frames=frames)

    # EXPORT
    def export(
        self,
        save_dirs=None,
        mode=None,
        frequency=30,
        metrics_to_export=list(PointMetrics),
        meter_per_pixel=None,
    ):
        human_dir = Path(save_dirs) / f"human_{self.human_idx}"
        for idx, (each, body_joint) in enumerate(self.body_joints.items()):
            body_joint.export(
                human_dir=human_dir,
                mode=mode,
                frequency=frequency,
                meter_per_pixel=meter_per_pixel,
                metrics_to_export=metrics_to_export,
            )

        for each, customRecord in self.customRecord.items():
            customRecord.export(
                human_dir=human_dir,
                mode=mode,
                frequency=frequency,
                meter_per_pixel=meter_per_pixel,
                metrics_to_export=metrics_to_export,
            )

    def export_metrics(self, dir):
        if self.human_idx:
            name = f"{self.human_idx}_{self.name}"
        else:
            name = self.name
        # export json
        json_output = self.get_metrics()
        json_file = Path(dir) / get_time()
        json_file = json_file / f"{name}.json"
        json_file.parent.mkdir(parents=True, exist_ok=True)
        with open(json_file, "w") as f:
            json.dump(json_output, f, indent=2)

    def export_csv(self, dir):
        # let's turn the data into csv file
        if self.human_idx:
            name = f"{self.human_idx}_{self.name}"
        else:
            name = self.name
        # First, let's get the metrics
        raw_result = self.get_metrics_flatten()
        df = pd.DataFrame(raw_result)
        csv_file = Path(dir) / get_time()
        csv_file = csv_file / f"{name}.csv"
        csv_file.parent.mkdir(parents=True, exist_ok=True)
        df.to_csv(csv_file)

    def get_metrics_flatten(self):
        # First, let's get the metrics
        metrics = self.get_metrics()
        # We need to flatten the metrics
        result = {}
        # First, let's get the metrics_info
        body_joints_metrics = metrics["body_joints_metrics"]

        for body, metric_data in body_joints_metrics.items():
            for metric_name, metric in metric_data.items():
                result[f"{body}_{metric_name}"] = metric

        custom_metrics = metrics["custom_metrics"]
        for body_part, metric_data in custom_metrics.items():
            for metric_name, metric in metric_data.items():
                result[f"{body_part}_{metric_name}"] = metric

        result["time"] = metrics["time"]
        return result
